import os
from os.path import dirname, abspath

from ..utils.helper import download_font_from_gdrive


class FontConfig:
    def __init__(self, font_type="terminal-grotesque-regular"):
        font_dir = dirname(abspath(__file__))
        self.font_dir = os.path.join(font_dir, "fonts")
        os.makedirs(self.font_dir, exist_ok=True)
        self.font_type = font_type
        self.download_font_file()

    @staticmethod
    def get_available_fonts():
        return ["terminal-grotesque-regular"]

    def get_font_file_location(self):
        if self.font_type == "terminal-grotesque-regular":
            return os.path.join(self.font_dir, "terminal-grotesque-regular.ttf")

    def download_font_file(self):
        if self.font_type == "terminal-grotesque-regular":
            path_to_download = os.path.join(self.font_dir, "terminal-grotesque-regular.ttf")
            file_id = "1UzXMTutTnfiDslCfQNKGWLdMlhi3KHiW"

            if not os.path.exists(path_to_download):
                if download_font_from_gdrive(file_id, path_to_download):
                    print(f"Downloaded font type {self.font_type}")
                else:
                    print(f"Download Failed")




from . config import CONFIGURATION
from . helper import generate_random_digits
from . helper import test_annotations
from . helper import format_annotations
from . helper import add_spaces
from . helper import generate_random_digits_with_probability
from . helper import generate_random_digits_with_positional_probability
from . helper import category_id
from . helper import convert_to_grayscale
from . helper import change_between_configs
from . helper import randomly_invert_config_color
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__all__ = [
    "CONFIGURATION",
    "generate_random_digits",
    "test_annotations",
    "format_annotations",
    "add_spaces",
    "generate_random_digits_with_probability",
    "generate_random_digits_with_positional_probability",
    "category_id",
    "convert_to_grayscale",
    "change_between_configs",
    "randomly_invert_config_color"
]


import random
import scipy.ndimage.filters as fi
import numpy as np

from .augmentation import SingleDigitAugmentation


class GaussianNoise(SingleDigitAugmentation):
    def __init__(self, mean=0, variance=0.1, probability=1):
        super(GaussianNoise, self).__init__(probability)
        self.mean = mean
        self.variance = variance

    def augment(self, image, annotation):
        row, col, ch = image.shape
        sigma = self.variance ** 0.5
        gauss = np.random.normal(self.mean, sigma, (row, col, ch))
        gauss = gauss.reshape(row, col, ch)
        noisy = image + gauss
        np.clip(noisy, 0, 255)
        return noisy, annotation


class PoissonNoise(SingleDigitAugmentation):
    def __init__(self, probability=1):
        super(PoissonNoise, self).__init__(probability)

    def augment(self, image, annotation):
        vals = len(np.unique(image))
        vals = 2 ** np.ceil(np.log2(vals))
        noisy = np.random.poisson(image * vals) / float(vals)
        np.clip(noisy, 0, 255)
        return noisy, annotation


class SPNoise(SingleDigitAugmentation):
    def __init__(self, s_vs_p=0.5, amount=0.004, probability=1):
        super(SPNoise, self).__init__(probability)
        self.s_vs_p = s_vs_p
        self.amount = amount

    def augment(self, image, annotation):
        out = np.copy(image)
        # Salt mode
        num_salt = np.ceil(self.amount * image.size * self.s_vs_p)
        coords = [np.random.randint(0, i - 1, int(num_salt))
                  for i in image.shape]
        coords = tuple(coords)
        out[coords] = 1

        # Pepper mode
        num_pepper = np.ceil(self.amount * image.size * (1. - self.s_vs_p))
        coords = [np.random.randint(0, i - 1, int(num_pepper))
                  for i in image.shape]
        coords = tuple(coords)
        out[coords] = 0
        np.clip(out, 0, 255)
        return out, annotation


class SpeckleNoise(SingleDigitAugmentation):
    def __init__(self, probability=1):
        super(SpeckleNoise, self).__init__(probability)

    def augment(self, image, annotation):
        row, col, ch = image.shape
        gauss = np.random.randn(row, col, ch)
        gauss = gauss.reshape(row, col, ch)
        noisy = image + image * gauss
        np.clip(noisy, 0, 255)
        return noisy, annotation


class ShadowPatch(SingleDigitAugmentation):
    def __init__(self, alpha=0.7, probability=1):
        super(ShadowPatch, self).__init__(probability)
        self.alpha = alpha

    def augment(self, image, annotation):
        t = random.randint(0, 2)
        if t == 1:
            image_size = image.shape
            kernel_size = (random.randint(0, image_size[0]), random.randint(0, image_size[1]))
            shadow = np.ones(kernel_size)
            x_displacement = random.randint(0, (image_size[0] - kernel_size[0]) // 2)
            x_remain = image_size[0] - kernel_size[0] - x_displacement
            y_displacement = random.randint(0, (image_size[1] - kernel_size[1]) // 2)
            y_remain = image_size[1] - kernel_size[1] - y_displacement
            shadow_patch = np.pad(shadow, ((x_displacement, x_remain), (y_displacement, y_remain)), 'constant',
                                  constant_values=[0]) * 255
            image = np.clip(image - (1 - self.alpha) * shadow_patch, 0, 255)
        return image, annotation


class FlashSpot(SingleDigitAugmentation):
    def __init__(self, probability=1, alpha=0.5):
        super(FlashSpot, self).__init__(probability)
        self.alpha = alpha

    @staticmethod
    def __gkernel(kernelLen=21, nsig=3):
        inp = np.zeros((kernelLen, kernelLen))
        kernel = fi.gaussian_filter(inp, nsig)
        return np.clip(kernel / kernel.max(), 0, 1) * 255

    def augment(self, image, annotation):
        t = random.randint(0, 2)
        if t == 1:
            kernel_size = random.randint(15, 35)
            nsig = random.randint(1, 7)
            image_size = image.shape
            flash = FlashSpot.__gkernel(kernel_size, nsig)
            x_displacement = random.randint(0, (image_size[0] - kernel_size) // 2)
            x_remain = image_size[0] - kernel_size - x_displacement
            y_displacement = random.randint(0, (image_size[1] - kernel_size) // 2)
            y_remain = image_size[1] - kernel_size - y_displacement
            flash_patch = np.pad(flash, ((x_displacement, x_remain), (y_displacement, y_remain)), 'constant',
                                 constant_values=[0])
            image = np.clip(image + self.alpha * flash_patch, 0, 255)
        return super().augment(image, annotation)



import unittest
import numpy as np
from digitgen import DigitGenerator


class TestGrayScale(unittest.TestCase):
    def test_default_grayscale_generation(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48), gray_scale=False)
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()

        self.assertTrue(ret_arr.shape == (100, 48, 128, 3))

        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48), gray_scale=True)
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()

        self.assertTrue(ret_arr.shape == (100, 48, 128))



import gdown
import numpy as np
from PIL import ImageDraw, Image
from numpy import random
import random as py_ran


def category_id(allowed_digits: list):
    return {k: v for k, v in enumerate(allowed_digits)}, {v: k for k, v in enumerate(allowed_digits)}


def randomly_insert(row, num_times, character):
    arr_char = np.full(num_times, character)
    tot_arr = np.concatenate((row, arr_char))
    random.shuffle(tot_arr)
    return tot_arr[:len(row)]


def randomly_invert_config_color(configs, probability, background_color, digit_color):
    n = 0
    for row in configs:
        ran_val = py_ran.random()

        if ran_val < probability:
            n += 1
            for config in row:
                config.background_color = background_color
                config.digit_color = digit_color

    print(f"num of inverts:-{n}")


def changing_width_constant_digits(configs, samples, digit_size, pixel_value):
    for row_id in range(samples):
        for each_id in range(digit_size - 1):
            configs[row_id][each_id].img_size[0] = configs[row_id][each_id].img_size[0] + pixel_value


def changing_width_random_digits(configs, samples, digit_size, pixel_range):
    for row_id in range(samples):
        for each_id in range(digit_size - 1):
            ran_num = py_ran.randint(pixel_range[0], pixel_range[1])
            configs[row_id][each_id].img_size[0] += ran_num


def change_between_configs(configs, mode, pixel_value, pixel_range, samples, digit_size):
    if mode == "constant":
        changing_width_constant_digits(configs, samples, digit_size, pixel_value)
    elif mode == "random":
        changing_width_random_digits(configs, samples, digit_size, pixel_range)
    else:
        pass


def convert_to_grayscale(arr, conv=None):
    if conv is None:
        conv = [0.2989, 0.5870, 0.1140]
    return np.dot(arr[..., :3], conv)


def split_insert(row, sectors, spaces_per_sector, digit_size, character):
    gap = digit_size // sectors
    for gap_num in range(1, sectors):
        for _ in range(spaces_per_sector):
            row = np.insert(row, gap_num * gap + (gap_num - 1) * spaces_per_sector, character)

    return row


def add_spaces(row, space_type, spaces, sectors, spaces_per_sector, digit_size):
    if space_type == "random":
        assert spaces
        row = randomly_insert(row, spaces, " ")

    if space_type == "space":
        assert sectors
        assert spaces_per_sector

        row = split_insert(row, sectors, spaces_per_sector, digit_size, " ")

    return row


def generate_random_digits_with_probability(digit_size, probability: dict):
    assert 0.999999999 < sum(probability.values()) < 1.000000000001

    allowed_digits = list(probability.keys())
    proba = list(probability.values())

    return np.random.choice(allowed_digits, size=digit_size, p=proba)


def generate_random_digits_with_positional_probability(digit_size, probability: dict):
    if digit_size != len(probability.keys()):
        assert "default" in probability.keys()

    __digits = []
    for digit_id in range(digit_size):
        try:
            prob = probability[digit_id]
        except KeyError:
            prob = probability["default"]

        __digits.append(generate_random_digits_with_probability(None, prob))

    return np.array(__digits)


def generate_random_digits(digit_size, allowed_digits):
    return np.random.choice(allowed_digits, size=digit_size)


def test_annotations(array, annotation):
    array = array.astype(np.uint8)
    image = Image.fromarray(array)
    draw = ImageDraw.Draw(image)

    for each in annotation:
        draw.rectangle(each["bbox"],
                       outline=tuple([255, 82, 82]),
                       width=1)

    image.show()


def format_annotations(full_ann, row_ann, category_map):
    if len(full_ann["annotations"]) == 0:
        current_id = 0
        current_img_id = 0
    else:
        current_id = full_ann["annotations"][-1]["id"] + 1
        current_img_id = full_ann["annotations"][-1]["image_id"] + 1

    for __id, ann in enumerate(row_ann):
        ann["id"] = current_id + __id
        ann["image_id"] = current_img_id

        full_ann["annotations"].append(ann)

    full_ann["category"] = category_map


def download_font_from_gdrive(__id, download_loc):
    try:
        gdown.download(id=__id, output=download_loc)
        return True
    except Exception as e:
        return False



import unittest
import numpy as np
from digitgen import DigitGenerator


class TestDigitSpaces(unittest.TestCase):
    def test_random_digit_generation(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        counts = np.unique(digit_gen.generated_digits, return_counts=True)[1]
        _max = max(counts)
        _min = min(counts)
        print(_max,_min)
        self.assertTrue(_max < 100 * 1.5 and _min > 100 * 0.5)

    def test_inline_probability(self):
        digit_gen = DigitGenerator(10, samples=1000, image_size=(128, 48))
        digit_gen.generate_digits_with_choice_probability({"0": 0.5, "1": 0.1, "2": 0.2, "3": 0.2})
        digit_gen.generate_digit_config()
        counts = np.unique(digit_gen.generated_digits, return_counts=True)[1]
        self.assertTrue(5000 * 1.2 > counts[0] > 5000 * 0.8)

    def test_position_probability(self):
        digit_gen = DigitGenerator(10, samples=1000, image_size=(128, 48))
        digit_gen.generate_digits_with_position_probability(
            {0: {"0": 0.5, "1": 0.1, "2": 0.2, "3": 0.2}, 1: {"4": 1}, 2: {"5": 0.5, "8": 0.5},
             "default": {"0": 0.1, "1": 0.1, "2": 0.1, "3": 0.1, "4": 0.1, "5": 0.1, "6": 0.1, "7": 0.1, "8": 0.1,
                         "9": 0.1}})
        digit_gen.generate_digit_config()
        for d in digit_gen.generated_digits:
            self.assertTrue(d[1] == "4")
            self.assertTrue(d[2] in ["5", "8"])



import unittest

from digitgen import DigitGenerator
from digitgen import test_annotations


class TestDigitSpaces(unittest.TestCase):
    def test_just_digits(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == 10)

    def test_space_sectors(self):
        digit_gen = DigitGenerator(12, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_spaces(space_type="space", sectors=3, spaces_per_sector=2)
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == 16)

    def test_space_random(self):
        digit_gen = DigitGenerator(10, samples=100, image_size=(128, 48))
        digit_gen.generate_digits()
        digit_gen.generate_spaces(space_type="random", num_spaces=3)
        digit_gen.generate_digit_config()
        ret_arr, ret_ann = digit_gen.generate_dataset()
        test_annotations(ret_arr[0], [x for x in ret_ann["annotations"] if x["image_id"] == 0])
        print([x["category_id"] for x in ret_ann["annotations"] if x["image_id"] == 0])
        self.assertTrue(len([x for x in ret_ann["annotations"] if x["image_id"] == 0]) == digit_gen.digit_size)
